# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Public API for tf.__internal__.tracking namespace.
"""

from __future__ import print_function as _print_function

import sys as _sys

from tensorflow.python.training.tracking.base import Trackable
from tensorflow.python.training.tracking.base import TrackableReference
from tensorflow.python.training.tracking.data_structures import TrackableDataStructure
from tensorflow.python.training.tracking.data_structures import sticky_attribute_assignment
from tensorflow.python.training.tracking.data_structures import wrap_or_unwrap as wrap
from tensorflow.python.training.tracking.tracking import AutoTrackable
from tensorflow.python.training.tracking.util import register_session_provider
from tensorflow.python.training.tracking.util import streaming_restore

del _print_function
