# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Public API for tf.__internal__.saved_model namespace.
"""

from __future__ import print_function as _print_function

import sys as _sys

from tensorflow.python.saved_model.load import load_partial
from tensorflow.python.saved_model.nested_structure_coder import StructureCoder

del _print_function
